﻿B4A=true
Group=Classes
ModulesStructureVersion=1
Type=Class
Version=9.8
@EndOfDesignText@
#RaisesSynchronousEvents: Parse
Sub Class_Globals
	Private moParser As SaxParser
	
	Type xtXmlElement (Name As String, Children As List, Text As String, Attributes As Map)
	
	Private moElements As List
 
End Sub


Public Sub Initialize
	moParser.Initialize
	moElements.Initialize
End Sub

' Returns a list of xtXmlElements
Public Sub XmlTree As List 
	Return moElements
End Sub

Public Sub TreeIsEmpty As Boolean
	Dim piCount As Int = 0
	If moElements <> Null And moElements.IsInitialized And moElements.Size > 0 Then 
		piCount = NodeCount(moElements.Get(0))
	End If
	Return (piCount < 2)
End Sub









Public Sub Parse(Xml As String) As Boolean 
	Dim In As InputStream
	Dim b() As Byte = Xml.GetBytes("UTF8")
	In.InitializeFromBytesArray(b, 0, b.Length)
	Return Parse2(In)
End Sub

Public Sub Parse2(Input As InputStream) As Boolean 
	moElements.Initialize
	moElements.Add(NewElement("root"))
	Try
		moParser.Parse(Input, "parser")
		Return True
	Catch
		LogColor("Xml2Map.Parse2: RTE " & LastException.Message, Colors.Red)
		moElements.Clear 
		Return False 
	End Try
End Sub

' Initializes internal tree with a "root" node and passes back that root node
Public Sub InitializeTree As xtXmlElement
	moElements.Initialize
	Dim poRoot As xtXmlElement = NewElement("root")
	moElements.Add(poRoot)
	Return poRoot 
End Sub


Public Sub GetNodeFromRoot(Name As String) As xtXmlElement
	If moElements.Size = 0 Then 
		Return Null
	Else
		Dim poRoot As xtXmlElement = moElements.Get(0)
		If Name = "root" Then
			Return poRoot
		Else
			Return GetNode(poRoot.Children,  Name)
		End If
	End If
End Sub

Public Sub GetNode(StartingNode As List, Name As String) As xtXmlElement
	Dim psCheck As String = Name.Trim.ToLowerCase
	If psCheck <> "" Then 
		For Each poEle As xtXmlElement In StartingNode
			If poEle.Name.Trim.ToLowerCase = psCheck Then 
				Return poEle
			Else
				If poEle.Children.Size > 0 Then 
					Dim poChildNode As xtXmlElement = GetNode(poEle.Children, Name)
					If poChildNode <> Null And poChildNode.IsInitialized Then 
						Return poChildNode
					End If
				End If
			End If
		Next
	End If
	Return Null
End Sub

Public Sub GetNodeTextFromRoot(Name As String) As String
	Dim psCheck As String = Name.Trim.ToLowerCase
	If psCheck <> "" Then
	Dim poNode As xtXmlElement = GetNodeFromRoot(Name)
		If poNode <> Null And poNode.IsInitialized Then 
			If poNode.Children.Size > 0 Then 
				Return GetNodeText(poNode.Children, Name)
			Else
				Return poNode.Text
			End If
		End If
	End If
	Return Null
End Sub

Public Sub GetNodeText(StartingNode As List, Name As String) As String
	Dim psCheck As String = Name.Trim.ToLowerCase
	If psCheck <> "" Then
		For Each poNode As xtXmlElement In StartingNode
			If poNode.Name.Trim.ToLowerCase = psCheck Then
				If poNode.Children.Size = 0 Then 
					Return poNode.Text
				Else
					Return GetNodeText(poNode.Children, Name)
				End If
			End If
		Next
	End If
	Return Null
End Sub


Public Sub GetNodeAttributesFromRoot(Name As String) As Map
	Dim psCheck As String = Name.Trim.ToLowerCase
	If psCheck <> "" Then
		Dim poNode As xtXmlElement = GetNodeFromRoot(Name)
		If poNode <> Null And poNode.IsInitialized Then 
			If poNode.Children.Size > 0 Then 
				Return GetNodeAttributes(poNode.Children, Name)
			Else
				Return poNode.Attributes
			End If
		End If
	End If
	Return Null 
End Sub

Public Sub GetNodeAttributes(StartingNode As List, Name As String) As Map 
	Dim psCheck As String = Name.Trim.ToLowerCase
	If psCheck <> "" Then
		For Each poNode As xtXmlElement In StartingNode
			If poNode.Name.Trim.ToLowerCase = psCheck Then
				If poNode.Children.Size = 0 Then 
					Return poNode.Attributes
				Else
					Return GetNodeAttributes(poNode.Children, Name)
				End If
			End If
		Next
	End If
	Return Null
End Sub

Public Sub NewElement(Name As String) As xtXmlElement
	Dim xe As xtXmlElement
	xe.Initialize
	xe.Children.Initialize
	xe.Name = Name
	xe.Attributes.Initialize
	xe.Text = ""
	Return xe
End Sub

'  Returns the current tree as an XML string, not including the "root" entry.
'  SoapEnvelope 1 = "soap:" tag, 2 = "soap12:" tag, anything else will omit soap envelope.
Public Sub ToXML(SoapEnvelope As Int, FormatForDisplay As Boolean) As String 
	Return BuildXMLFromNode("root", SoapEnvelope, FormatForDisplay).ToString 
End Sub

'  SoapEnvelope 1 = "soap:" tag, 2 = "soap12:" tag, anything else will omit soap envelope.
Public Sub ToXMLFromNode(NodeName As String, SoapEnvelope As Int, FormatForDisplay As Boolean) As String 
	Return BuildXMLFromNode(NodeName, SoapEnvelope, FormatForDisplay).ToString 
End Sub

'  Dumps the current tree from the root node 
Public Sub DebugWalkTree
	If moElements <> Null And moElements.IsInitialized And moElements.Size > 0 Then
		LogColor("clsXmlTree: Begin tree data -----------" & CRLF, Colors.RGB(80, 80, 184))
		Dim poNode As xtXmlElement = moElements.Get(0)
		DebugWalkNode(poNode.Children, "")
		LogColor("clsXmlTree: End tree data -----------" & CRLF, Colors.RGB(80, 80, 184))
	Else
		LogColor("clsXmlTree: No current tree data.", Colors.Red)
	End If
	
End Sub





'  Returns an XML string starting with (but not including) the supplied node name.
'  SoapEnvelope = 1 for "soap:" or 2 for "soap12:" tags.  Anything else will omit soap envelope.
Private Sub BuildXMLFromNode(NodeName As String, SoapEnvelope As Int, MakeItPretty As Boolean) As StringBuilder
	Dim poSB As StringBuilder
	Dim poStart As xtXmlElement
	
	poSB.Initialize
	poStart = GetNodeFromRoot(NodeName)
	If poStart <> Null And poStart.IsInitialized Then
		' skip the starting element by starting with it's children
		Dim psIndent As String = ""
		If MakeItPretty Then
			psIndent = "      "
		End If
		BuildWalkXmlNode(poStart.Children, poSB, MakeItPretty, psIndent)
	End If
	Select Case SoapEnvelope
		Case 1
			WrapSoapEnvelope(poSB, MakeItPretty)
		Case 2
			WrapSoap12Envelope(poSB, MakeItPretty)

		Case Else
			WrapXmlHeader(poSB, MakeItPretty)
	End Select
	Return poSB
End Sub


Private Sub BuildWalkXmlNode(oNode As List, oXmlString As StringBuilder, MakeItPretty As Boolean, CurrentIndent As String)
	For Each poCurrNode As xtXmlElement In oNode
		oXmlString.Append(CurrentIndent).Append("<").Append(XmlEncode(poCurrNode.Name))
		If poCurrNode.Attributes.IsInitialized And poCurrNode.Attributes.Size > 0 Then
			For Each psKey As String In poCurrNode.Attributes.Keys
				Dim psVal As String = poCurrNode.Attributes.Get(psKey)
				oXmlString.Append($" ${XmlEncode(psKey)}="${XmlEscape(psVal)}""$)
			Next
		End If
		If poCurrNode.Children.Size > 0 Then
			oXmlString.Append(">")
			Dim psDeeper As String
			If MakeItPretty Then
				psDeeper = CurrentIndent & "   "
				oXmlString.Append(CRLF)
			Else
				psDeeper = CurrentIndent
			End If
			BuildWalkXmlNode(poCurrNode.Children, oXmlString, MakeItPretty, psDeeper)
			oXmlString.append(CurrentIndent).Append("</").Append(XmlEncode(poCurrNode.Name)).Append(">")
			If MakeItPretty Then
				oXmlString.Append(CRLF)
			End If
		Else
			If poCurrNode.Text.Trim = "" Then
				oXmlString.Append(" />")
			Else
				oXmlString.Append(">").Append(XmlEscape(poCurrNode.Text))
				oXmlString.Append("</").Append(XmlEncode(poCurrNode.Name)).Append(">")
			End If
			If MakeItPretty Then
				oXmlString.Append(CRLF)
			End If
		End If
	Next
End Sub


Private Sub XmlEncode(sText As String) As String
	Return sText.Replace("<", "").Replace(">", "").Replace("/", "").Replace(QUOTE, "").Replace("&", "").Replace("'", "")
End Sub

Private Sub XmlEscape(sText As String) As String
	Return sText.Replace("<", "&lt;").Replace(">", "&gt;").Replace(QUOTE, "&quot;").Replace("&", "&amp;").Replace("'", "&apos;")
End Sub

'  Puts outer XML header  (envelope and body) around the meat of the XML 
Private Sub WrapSoapEnvelope(InputXML As StringBuilder, MakeItPretty As Boolean)
	Dim psXML As String = $"<?xml version="1.0" encoding="utf-8"?>"$
	Dim psHeader As String = $"<soap:Envelope xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/"><soap:Body>"$
	Dim psFooter As String = $"</soap:Body></soap:Envelope>"$
	If MakeItPretty Then
		psXML = psXML & CRLF 
		psHeader = psHeader & CRLF
		psFooter =  CRLF & psFooter & CRLF
	End If
	InputXML.Insert(0, psXML)
	InputXML.Insert(psXML.Length, psHeader)
	InputXML.Append(psFooter)
End Sub

'  Puts outer XML header  (envelope and body) around the meat of the XML 
Private Sub WrapSoap12Envelope(InputXML As StringBuilder, MakeItPretty As Boolean)
	Dim psXML As String = $"<?xml version="1.0" encoding="utf-8"?>"$
	Dim psHeader As String = $"<soap12:Envelope xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:soap12="http://www.w3.org/2003/05/soap-envelope"><soap12:Body>"$
	Dim psFooter As String = $"</soap12:Body></soap12:Envelope>"$
	If MakeItPretty Then
		psXML = psXML & CRLF 
		psHeader = psHeader & CRLF
		psFooter =  CRLF & psFooter & CRLF
	End If
	InputXML.Insert(0, psXML)
	InputXML.Insert(psXML.Length, psHeader)
	InputXML.Append(psFooter)
End Sub


Private Sub WrapXmlHeader(InputXML As StringBuilder, MakeItPretty As Boolean)
	Dim psHeader As String = $"<?xml version="1.0" encoding="utf-8"?>"$
	If MakeItPretty Then
		psHeader = psHeader & CRLF
	End If
	InputXML.Insert(0, psHeader)
End Sub

Private Sub NodeCount(CurrentNode As xtXmlElement) As Int 
	Dim piCount As Int 
	
	piCount = piCount + 1
	If CurrentNode.Children.IsInitialized And CurrentNode.Children.Size > 0 Then 
		For Each poChildNode As xtXmlElement In CurrentNode.Children
			piCount = piCount + NodeCount(poChildNode)
		Next
	End If
	Return piCount 	
End Sub

Private Sub DebugWalkNode(oNode As List, sIndent As String)
	For Each poNode As xtXmlElement In oNode
		LogColor(sIndent & poNode.Name, Colors.Magenta)
		If poNode.Children.Size > 0 Then
			Dim psDeeper As String = sIndent & "---+"
			DebugWalkNode(poNode.Children, psDeeper)
		Else
			LogColor(sIndent & poNode.Name & " = " & poNode.Text, Colors.RGB(80, 80, 244))
		End If
	Next
End Sub


Private Sub Parser_StartElement (Uri As String, Name As String, Attributes As Attributes)
	Dim poElement As xtXmlElement = NewElement(Name)
	If Attributes.IsInitialized And Attributes.Size > 0 Then
		Dim piIndex As Int 
		Dim poMap As Map 
		poMap.Initialize
		For piIndex = 0 To Attributes.Size - 1
			poMap.Put(Attributes.GetName(piIndex), Attributes.GetValue(piIndex))
		Next
		poElement.Attributes = poMap 
	End If
	GetLastElement.Children.Add(poElement)
	moElements.Add(poElement)
End Sub

Private Sub Parser_EndElement (Uri As String, Name As String, Text As StringBuilder)
	Dim poElement As xtXmlElement = GetLastElement
	poElement.Text = Text.ToString
	moElements.RemoveAt(moElements.Size - 1)
End Sub

Private Sub GetLastElement As xtXmlElement
	Return moElements.Get(moElements.Size - 1)
End Sub










